"""
CLI module for NoPrint
"""
import sys
import argparse
import contextvars
from multiprocessing import cpu_count

import noprint.logger as logging

from noprint.sprint import detect_prints


log_lvl = contextvars.ContextVar("log_lvl")
error_out = contextvars.ContextVar("error_out")


def parse_args():
    """No prints are allowed!"""
    parser = argparse.ArgumentParser(
        prog="NoPrint",
        description="Do not allow prints in your code.",
        epilog="Thank you for using NoPrint",
        allow_abbrev=False,
    )
    parser.add_argument(
        "-e",
        "--error-out",
        action="store_true",
        help="exit with error when print is found (by default only warnings are shown)",
    )
    parser.add_argument(
        "-f", "--first-only", action="store_true", help="finish on first print found"
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="provide more analysis information (use multiple v's to increase logging level)",
    )
    parser.add_argument(
        "packages",
        help="which packages/modules to check, syntax: <package>[.<module> ...], e.g. noprint or noprint.cli",
        nargs="+",
        type=str,
    )
    parser.add_argument(
        "-m",
        "--multi",
        nargs="?",
        const=1,
        type=int,
        help="set how many threads to use",
    )
    parser.add_argument("--version", action="version", version="%(prog)s 3.1.0")
    args = parser.parse_known_intermixed_args()[0]

    err_out = args.error_out
    first_only = args.first_only
    verbose = bool(args.verbose)
    very_verbose = args.verbose >= 2
    packages = args.packages
    multi = (
        cpu_count()
        if args.multi is not None and args.multi <= 0
        else args.multi
        if args.multi
        else 1
    )  # cpu_count when <=0; 1 when not given; otherwise multi

    lvl = logging.ERROR if err_out else logging.WARNING

    var = contextvars.ContextVar("packages")
    var.set(packages)
    var = contextvars.ContextVar("first_only")
    var.set(first_only)
    var = contextvars.ContextVar("verbose")
    var.set(verbose)
    var = contextvars.ContextVar("very_verbose")
    var.set(very_verbose)
    var = contextvars.ContextVar("mt_threads")
    var.set(multi)

    log_lvl.set(lvl)
    error_out.set(err_out)


def cli():
    """CLI function"""
    parse_args()

    result = detect_prints()

    if result == 2:
        logging.log("Exiting with critical status", logging.CRITICAL)
    elif result == 1:
        logging.log("Print statements detected", log_lvl.get())
    else:
        logging.log("No print statements found, cheers 🍺", logging.INFO)

    if not error_out.get() and result == 1:
        result = 0
    sys.exit(result)  # 0 when success, 1 for error, 2 for critical



"""
Module with tests for noprint.module
"""
from pathlib import Path
from unittest import mock
from importlib.machinery import ModuleSpec

import pytest

from noprint.exceptions import ParentModuleNotFoundException
from noprint.module import (
    Module,
    _get_module_search_path,
    _get_module_location,
    _find_parent_dir,
    _next_path,
    _package_to_dir,
)


@pytest.mark.parametrize("search_loc", [["/root"], mock.Mock()])
def test__get_module_search_path(search_loc):
    """Testing _get_module_search_path"""
    mspec = mock.Mock()
    if isinstance(search_loc, mock.Mock):
        search_loc._path = ["/root"]  # pylint:disable=protected-access
    mspec.submodule_search_locations = search_loc
    assert _get_module_search_path(mspec) == "/root"


@pytest.mark.parametrize("has_loc", [False, True])
@pytest.mark.parametrize("file_name", ["file.py", "__init__.py", "file.nopy"])
def test__get_module_location(has_loc, file_name):
    """Testing _get_module_location"""
    mspec = mock.Mock()
    mspec.has_location = has_loc
    mspec.origin = f"/root/test/{file_name}"
    with mock.patch(
        "noprint.module._get_module_search_path", side_effect=["/root/test"]
    ):
        if not has_loc or file_name == "__init__.py":
            assert _get_module_location(mspec) == str(Path("/root"))
        else:
            assert _get_module_location(mspec) == str(Path("/root/test"))


@pytest.mark.parametrize("in_cwd", [False, True])
@pytest.mark.parametrize("found", [False, True])
def test__find_parent_dir(in_cwd, found):
    """Testing _find_parent_dir"""
    with mock.patch("noprint.module.PathFinder.find_spec") as mod_findspec:
        mock_mod = mock.Mock(spec=ModuleSpec)
        mock_mod.origin = "origin"
        mod_findspec.side_effect = [mock_mod if found else None]
        if not found:
            with pytest.raises(ParentModuleNotFoundException):
                _find_parent_dir("test", in_cwd)
        else:
            res = _find_parent_dir("test", in_cwd)
            assert res == "."


@pytest.mark.parametrize(
    "isdir", [[True], [False, True], [False, False, True], [False, False, False]]
)
def test__next_path(isdir):
    """Testing _next_path"""
    with mock.patch("noprint.module.os.path.isdir", side_effect=isdir):
        submodules = "test.test.test"
        res = _next_path("test", submodules)
        if any(isdir):
            count = sum(not bls for bls in isdir)
            assert res == str(Path("test/" + submodules.rsplit(".", maxsplit=count)[0]))
        else:
            assert res is None


def test__package_to_dir():
    """Testing _package_to_dir"""
    with mock.patch("noprint.module.os.path.isdir", return_value=True):
        submodules = "test.test.test"
        assert str(Path(_package_to_dir("test", submodules))) == str(
            Path("test/test.test.test/")
        )
    with mock.patch("noprint.module.os.path.isdir", return_value=False):
        submodules = "test.test.test"
        assert _package_to_dir("test", submodules) is None


@pytest.mark.parametrize("found", [False, True])
def test_module___init__(found):
    """Testing Module.__init__"""
    with mock.patch(
        "noprint.module._find_parent_dir",
        side_effect=["/root" if found else Exception("exc")],
    ):
        if found:
            mod = Module(package="mod.module")
            assert mod.name == "mod.module"
            assert mod._parent_loc == "/root"  # pylint:disable=protected-access
        else:
            with pytest.raises(ParentModuleNotFoundException):
                Module(package="mod.module")


@pytest.mark.parametrize("isdir", [False, True])
@pytest.mark.parametrize("isfile", [False, True])
@mock.patch("noprint.module.os.path.isdir")
@mock.patch("noprint.module.os.path.isfile")
def test_module_origin(mock_isfile, mock_isdir, mock_module, isdir, isfile):
    """Testing Module.origin"""
    mock_isdir.return_value = isdir
    mock_isfile.return_value = isfile
    module = mock_module()
    if isfile is False:
        assert module.origin == []
    elif isdir is False:
        assert module.origin == [str(Path("/root/test/subpackage.py"))]
    else:
        assert "__init__.py" in [origin[-11:] for origin in module.origin]
        assert "__main__.py" in [origin[-11:] for origin in module.origin]


@pytest.mark.parametrize("isdir", [False, True])
@mock.patch("noprint.module.os.path.isdir")
def test_module_search_path(mock_isdir, mock_module, isdir):
    """Testing Module.search_path"""
    mock_isdir.return_value = isdir
    module = mock_module()
    with mock.patch("noprint.module._package_to_dir", return_value="/root"):
        assert [module.search_path] == ["/root" if isdir else None]


def test_module___eq__(mock_module):
    """Testing Module.__eq__"""
    module = mock_module()
    module_other = mock_module()
    assert module == module_other
    module_other._parent_loc = "/nonroot"  # pylint:disable=protected-access
    assert module != module_other
    assert module != 0



"""
Logging setup for NoPrint
"""

import os
import sys

from pathlib import Path
from functools import lru_cache
from importlib.machinery import ModuleSpec, PathFinder

from noprint.exceptions import ParentModuleNotFoundException


def _get_module_search_path(module: ModuleSpec):
    """Get path where submodules are located"""
    pkg_path = module.submodule_search_locations
    if isinstance(pkg_path, list):
        pkg_path = pkg_path[0]
    else:  # pragma: no cover
        # Patch for _NamespacePath in Python3.7
        pkg_path = pkg_path._path[0]  # pylint: disable=protected-access
    return pkg_path


def _get_module_location(module: ModuleSpec):
    """Get location of the module"""
    if module.has_location:
        path = Path(module.origin)
        if path.name == "__init__.py":
            path = path.parents[1]
        else:
            path = path.parents[0]
    else:
        path = Path(_get_module_search_path(module))
        path = path.parent
    return str(path)


@lru_cache(maxsize=None)
def _find_parent_dir(package: str, in_cwd: bool) -> str:
    """Get location of submodule"""
    paths = sys.path.copy()
    if in_cwd:
        paths.insert(0, os.getcwd())

    module = PathFinder.find_spec(package, path=paths)

    if module is None:
        raise ParentModuleNotFoundException("Parent module not found")
    path = _get_module_location(module)

    return path


def _next_path(path, submodules):
    for i, _ in enumerate(submodules.split(".")):
        dir_check = submodules.rsplit(".", maxsplit=i)[0]
        dir_check = os.path.join(path, dir_check)
        if os.path.isdir(dir_check):
            return dir_check
    return None


def _package_to_dir(path, submodules):
    package_path = None
    for _ in enumerate(submodules.split(".")):
        package_path = _next_path(path, submodules)
        if package_path:
            submodules = submodules[len(package_path) - len(path) :]
            if os.path.isdir(package_path):
                path = package_path
        else:
            return None
    return package_path


class Module:
    """Module class to use instead of official classes which import main package when submodule is provided"""

    _origin = None
    _name = None

    _parent_loc = None  # Location of main parent pacakge (e.g. main.submodule.sm will return main)
    _package = None

    _search_path = None

    def __init__(self, package: str, in_cwd: bool = True):
        self._package = package
        try:
            ppackage = package.split(".", maxsplit=1)[0]
            self._parent_loc = _find_parent_dir(package=ppackage, in_cwd=in_cwd)
        except Exception as exc:
            raise ParentModuleNotFoundException(exc) from exc

    @property
    def origin(self):
        """Function to recover module origin - file paths"""
        if self._origin is None:
            path = Path(self._parent_loc)
            for step in self._package.split(".")[0:-1]:
                path = os.path.join(path, step)
            # SRC path
            cur_package = self._package.rsplit(".", maxsplit=1)[-1]
            self._origin = []

            # Check if module with submodules
            if os.path.isdir(os.path.join(path, cur_package)):
                path = os.path.join(path, cur_package)
                candidates = ["__init__.py", "__main__.py"]
                for candidate in candidates:
                    if os.path.isfile(os.path.join(path, candidate)):
                        self._origin.append(os.path.join(path, candidate))
            else:
                if os.path.isfile(os.path.join(path, f"{cur_package}.py")):
                    self._origin.append(os.path.join(path, f"{cur_package}.py"))
        return self._origin

    @property
    def name(self):
        """Recover package name"""
        return self._package

    @property
    def search_path(self):
        """Get path to use for searching submodules"""
        if self._search_path is None:
            self._search_path = _package_to_dir(self._parent_loc, self._package)
        if self._search_path and os.path.isdir(self._search_path):
            return self._search_path
        return None

    def __eq__(self, other):
        if isinstance(other, Module):
            pkg_check = self.name == other.name
            ppath_check = self._parent_loc == other._parent_loc
            return all([pkg_check, ppath_check])
        return False



"""
Module with tests for noprint.logger
"""
import logging

from unittest import mock

import pytest

import noprint.logger


@pytest.mark.parametrize(
    "lvl", [logging.INFO, logging.WARNING, logging.ERROR, logging.CRITICAL]
)
@mock.patch("noprint.logger.logger.info")
@mock.patch("noprint.logger.logger.warning")
@mock.patch("noprint.logger.logger.error")
@mock.patch("noprint.logger.logger.critical")
def test_log(mock_c, mock_e, mock_w, mock_i, lvl):
    """Function for testing _log method"""
    noprint.logger.log("testmsg", lvl)
    if lvl == logging.CRITICAL:
        mock_c.assert_called_once()
    elif lvl == logging.ERROR:
        mock_e.assert_called_once()
    elif lvl == logging.WARNING:
        mock_w.assert_called_once()
    elif lvl == logging.INFO:
        mock_i.assert_called_once()



"""
Module with unit tests for noprint.cli
"""
from unittest import mock

import pytest

import noprint.cli


@pytest.mark.parametrize("detected", [0, 1, 2])
@pytest.mark.parametrize("as_error", [0, 1])
@pytest.mark.parametrize("verbosity", [0, 1, 2])
@mock.patch("noprint.cli.detect_prints")
@mock.patch("noprint.cli.logging")
def test_cli(
    mock_log, mock_detect, verbosity, as_error, detected
):  # pylint: disable=unused-argument
    """Function for testing cli method"""
    mock_detect.return_value = detected

    args = ["noprint", "noprint", "test"]
    if as_error:
        args.append("-e")
    for _ in range(verbosity):
        args.append("-v")

    with mock.patch("sys.argv", args), pytest.raises(SystemExit) as syse:
        noprint.cli.cli()
    assert syse.type == SystemExit
    if detected == 2:
        assert syse.value.code == 2
    else:
        assert bool(syse.value.code) is bool(as_error and detected)



"""
Module with tests for noprint.sprint
"""
import contextvars
from unittest import mock

import pytest

import noprint

from noprint.sprint import (
    PackageFinder,
    _parse_pyfile,
    _get_module,
    _get_subpackages,
    _parse_module,
)
from noprint.exceptions import ImportException, ParentModuleNotFoundException


@pytest.mark.parametrize("origin", [None, "origin", "__init__.py"])
@pytest.mark.parametrize("name", ["name", "__pycache__"])
@mock.patch("noprint.sprint.os")
def test_get_subpackages(mock_os, origin, name):
    """Testing function for _get_subpackages - mock several different module specs and finish with a proper one"""
    mod_mock = mock.MagicMock()
    mod_mock.origin = [origin] if origin is not None else []
    mod_mock.name = name
    mod_mock.search_path = "loc"
    mod_mock.submodule_search_locations = ["loc"]

    mod_mock_e = mock.MagicMock()
    mod_mock_e.origin = ["origin"]
    mod_mock_e.name = "name"
    mod_mock_e.search_path = "loc"
    mod_mock_e.submodule_search_locations = ["loc"]

    mock_os.listdir.return_value = ["name"]

    for package in _get_subpackages("noprint", module=mod_mock):
        assert package is not None


@pytest.mark.parametrize("sub_pkgs", [[], [mock.Mock()], [mock.Mock(), mock.Mock()]])
@pytest.mark.parametrize("origin", [None, "origin"])
def test_parse_module(origin, sub_pkgs):
    """Testing function for _get_subpackages - mock several different module specs and finish with a proper one"""
    module = mock.Mock()
    module.origin = origin
    with mock.patch("noprint.sprint._get_module", return_value=module), mock.patch(
        "noprint.sprint._get_subpackages", return_value=sub_pkgs
    ):
        result = _parse_module(package="noprint")
        if module.origin:
            assert result[0] == module
        else:
            assert result[0] is None
        assert len(result[1]) == len(sub_pkgs)


def _parse_mock(package):  # pylint:disable=unused-argument
    """Helper function for mocking _parse_module (Pool pickling)"""
    return (package, [])


def test_pf_packages_iter():
    """Testing PackageFinder.packages_iter"""
    pkg_finder = PackageFinder()

    noprint.sprint._parse_module = _parse_mock  # pylint:disable=protected-access

    packages = ("source", "test")
    var = contextvars.ContextVar("packages")
    var.set(packages)
    noprint.sprint.packages = var
    assert set(pkg_finder.packages_iter()) == set(packages)


@pytest.mark.parametrize(
    "ret",
    [
        [mock.MagicMock(), mock.MagicMock()],
        [mock.MagicMock(), None],
        [mock.MagicMock()] * 2,
    ],
)
def test__get_module__mod(ret):
    """Testing function for _get_subpackages - mock several different module specs and finish with a proper one"""

    with mock.patch("noprint.sprint.Module") as mock_mod:
        if isinstance(ret[1], mock.MagicMock):
            ret[1].origin.__len__.return_value = 1
        mock_mod.side_effect = ret
        _get_module("noprint")


@pytest.mark.parametrize(
    "ret", [None, ParentModuleNotFoundException("Dummy exception")]
)
def test__get_module__import_exc(ret):
    """Testing function for _get_subpackages - mock several different module specs and finish with a proper one"""

    with mock.patch("noprint.sprint.Module") as mock_mod, pytest.raises(
        ImportException
    ):
        if ret is None:
            mock_mod.return_value = ret
        else:
            mock_mod.side_effect = ret
        _get_module("noprint")


@pytest.mark.parametrize("code", ["print('')", "", "i=1"])
@pytest.mark.parametrize("mod", [None, ImportException("X")])
@mock.patch("builtins.open")
def test__parse_pyfile(mock_open, mod, code):
    """Test method for _parse_python - finding print statements"""
    fin = mock.Mock()
    fin.return_value.readline.return_value = "# -*- coding: utf-8-sig -*-"
    fin.return_value.read.return_value = code
    mock_open.return_value.__enter__ = fin

    module = mock.Mock()
    module.origin = ["noprint"]
    module.name = "noprint"

    if mod is None:
        mod = module

    res = _parse_pyfile(mod)

    if isinstance(mod, ImportException):
        assert res == 2
    elif "print" in code:
        assert res == 1  # ([("[noprint] Line: 1", True)], None)
    else:
        assert res == 0  # ([("[CLEAR]:[noprint]", False)], None)


@pytest.mark.parametrize("first", [True, False])
@pytest.mark.parametrize(
    "mod",
    [
        (
            [("X", False)],
            None,
        ),
        (
            [("X", True)],
            None,
        ),
        (
            [],
            ImportException("X"),
        ),
    ],
)
@mock.patch("noprint.sprint._parse_pyfile", side_effect=lambda *args, **kwargs: 0)
def test_pf_run(mock_parse, mod, first):  # pylint: disable=unused-argument
    """Testing function for _get_prints - verifying if code contains print statements"""

    def _subpackages():  # pylint: disable=unused-argument
        i = 0
        yield mod
        i = i + 1
        if i == 10:
            return

    with mock.patch(
        "noprint.sprint.PackageFinder.packages_iter", side_effect=_subpackages
    ):
        result = PackageFinder().run()
        packages = ("source", "test")
        var = contextvars.ContextVar("packages")
        var.set(packages)
        noprint.sprint.packages = var
        assert result == 0



"""
Logging setup for NoPrint
"""
import logging

from logging import INFO, WARNING, ERROR, CRITICAL


def log(msg: str, level: int):
    """Print with error or warning styling"""
    if level == INFO:
        logger.info(msg)
    elif level == WARNING:
        logger.warning(msg)
    elif level == ERROR:
        logger.error(msg)
    elif level == CRITICAL:
        logger.critical(msg)


logger = logging.getLogger("noprint")

formatter = logging.Formatter("%(levelname)s:%(message)s")
formatter_success = logging.Formatter("\033[1;32m%(message)s\033[1;0m")

ch = logging.StreamHandler()
ch.setLevel(WARNING)
ch.setFormatter(formatter)

chs = logging.StreamHandler()
chs.setLevel(INFO)
chs.setFormatter(formatter_success)
chs.semit = chs.emit
chs.emit = lambda record: chs.semit(record) if record.levelno == chs.level else None

logger.addHandler(ch)
logger.addHandler(chs)
logger.setLevel(logging.INFO)

logging.addLevelName(WARNING, f"\033[1;33m[{logging.getLevelName(WARNING)}]\033[1;0m")
logging.addLevelName(ERROR, f"\033[1;31m[{logging.getLevelName(ERROR)}]\033[1;0m")
logging.addLevelName(
    CRITICAL, f"\033[1;101m[{logging.getLevelName(CRITICAL)}]\033[1;0m"
)
# LogLevel.INFO doesn't display levelname, so not needed



"""
Exceptions module
"""


class ImportException(Exception):
    "Raised when there was import exception in print_seeker"


class ParentModuleNotFoundException(Exception):
    "Raised when there was import exception in print_seeker"



"""
Module responsible for finding print statements in python modules
"""
import os
import re
import ast
import pkgutil
import functools
import contextvars
from pathlib import Path
from multiprocessing.pool import Pool

import noprint.logger as logging

from noprint import ENCODING_CAPTURE
from noprint.module import Module
from noprint.exceptions import ImportException, ParentModuleNotFoundException


packages = contextvars.ContextVar("packages", default=[])
mt_threads = contextvars.ContextVar("mt_threads", default=1)
log_lvl = contextvars.ContextVar("log_lvl", default=logging.WARNING)
first_only = contextvars.ContextVar("first_only", default=False)
verbose = contextvars.ContextVar("verbose", default=False)
very_verbose = contextvars.ContextVar("very_verbose", default=False)


def _get_module(package):
    """Get Module of the package"""
    try:
        module = Module(package)
    except ParentModuleNotFoundException as exc:
        raise ImportException(exc) from exc
    if not module:
        raise ImportException(
            f"Module [{package}] is not present in current environment, directory or PYTHONPATH"
        )

    system_module = None
    try:
        system_module = Module(package, in_cwd=False)
    except ParentModuleNotFoundException:  # pragma: no cover
        pass

    if not system_module and verbose.get():
        logging.log(
            f"Module [{package}] is not installed", logging.WARNING
        )  # pragma: no cover
    elif (
        verbose.get()
        and system_module
        and module != system_module
        and len(system_module.origin) > 0
    ):  # pragma: no cover
        logging.log(
            f"Module [{package}] is overshadowing installed module", logging.WARNING
        )

    return module


def _get_subpackages(package, module):
    """Get all candidates for submodules"""
    # If module is a file or contains __init__ then yield it and set flag
    isinit = False
    if module.origin:
        candidates = [Path(orig).name for orig in module.origin]
        isinit = "__init__.py" in candidates

    # Grab submodules and all packages within the directory
    candidates = []
    sub_pkgs = []
    if not module.origin or isinit:
        pkg_path = module.search_path
        if pkg_path:  # Found non-.py file as submodule
            candidates = [
                f"{module.name}.{name}"
                for name in os.listdir(pkg_path)
                if os.path.isdir(os.path.join(pkg_path, name)) and name != "__pycache__"
            ]
            sub_pkgs = [
                ".".join([package, name])
                for _, name, _ in pkgutil.iter_modules([pkg_path])
            ]
    # If submodule is a directory and doesn't contain __init__ raise Warning
    candidates_missing = set(candidates) - set(sub_pkgs)
    if isinit and candidates_missing and verbose.get():  # pragma: no cover
        for candidate in candidates_missing:
            logging.log(f"Module [{candidate}] has no __init__.py", logging.WARNING)
    # Patch missing submodules
    return list(set(candidates) | set(sub_pkgs))


def _parse_module(package: str):
    """Grab all packages and subpackages"""
    module = _get_module(package)

    sub_pkgs = _get_subpackages(package, module)

    funcs = []
    for pkg in sub_pkgs:
        funcs = funcs + [functools.partial(_parse_module, package=pkg)]
    return module if module.origin else None, funcs


def _parse_pyfile(module):
    """Method for parsing python source code files to look for prints"""
    if isinstance(module, ImportException):
        logging.log(module.args[0], logging.CRITICAL)
        return 2

    encoding = "utf-8"
    status = 0
    # First two lines of Python source code have to be ASCII compatible
    # PEP-8, PEP-263, PEP-3120
    for mod_file in module.origin:
        with open(mod_file, "r", encoding="utf-8") as file:
            for _ in range(2):  # Check 1st two lines
                try:
                    found = re.search(ENCODING_CAPTURE, file.readline())
                # pylint: disable=broad-exception-caught
                except Exception as exc:  # pragma: no cover
                    logging.log(exc.args[0], logging.CRITICAL)
                    return 2
                if found:
                    encoding = found.group(1)
                    break

        with open(mod_file, "r", encoding=encoding) as file:
            parsed = ast.parse(file.read())
            clear = True
            name = ""
            if mod_file.endswith("__init__.py") or mod_file.endswith("__main__.py"):
                name = f".{mod_file[-11:-3]}"  # pragma: no cover
            for node in ast.walk(parsed):
                if node.__dict__.get("id") == "print":
                    clear = False
                    status = 1
                    if verbose.get():  # pragma: no cover
                        logging.log(
                            f"[{module.name}{name}] Line: {node.lineno}", log_lvl.get()
                        )

                    if first_only.get():  # pragma: no cover
                        return status
            if clear and very_verbose.get():
                logging.log(
                    f"[CLEAR]:[{module.name}{name}]", logging.INFO
                )  # pragma: no cover
    return status


class PackageFinder:
    """Class responsible for finding all packages and handling multiprocessing"""

    processes = []
    results = []

    def __init__(self):  # pragma: no cover
        self.pool = Pool(mt_threads.get())

    def err_callback(self, exc):  # pragma: no cover
        """Error callback from searching module"""
        self.results.append(exc)

    def callback(self, res):  # pragma: no cover
        """Callback from finding module"""
        self.results.append(res[0])
        for func in res[1]:
            proc = self.pool.apply_async(
                func, callback=self.callback, error_callback=self.err_callback
            )
            self.processes.append(proc)

    def packages_iter(self):
        """Iterate over all provided subpackages"""
        for package in packages.get():
            func = functools.partial(_parse_module, package=package)
            proc = self.pool.apply_async(
                func, callback=self.callback, error_callback=self.err_callback
            )
            self.processes.append(proc)

        while len(self.processes) > 0:
            for job in list(self.processes):
                if job.ready():
                    self.processes.remove(job)
                    res = self.results.pop()
                    if res:
                        yield res

    def run(self):
        """Find print statements and potential exceptions from selected packages"""
        results = []

        for module in self.packages_iter():
            res = _parse_pyfile(module)
            results.append(res)
            if res >= 1 and first_only.get():  # pragma: no cover
                break
        self.pool.terminate()
        self.pool.join()
        return max(results)


def detect_prints() -> int:  # pragma: no cover
    """Detect print statements from packages found by _get_subpackages"""
    # pylint: disable=unnecessary-lambda-assignment
    get_var = lambda name, ctx: [var.get() for var in iter(ctx) if var.name == name][0]

    ctx = contextvars.copy_context()

    packages.set(get_var("packages", ctx))
    mt_threads.set(get_var("mt_threads", ctx))
    log_lvl.set(get_var("log_lvl", ctx))
    first_only.set(get_var("first_only", ctx))
    verbose.set(get_var("verbose", ctx))
    very_verbose.set(get_var("very_verbose", ctx))

    logging.log(
        "Starting analysis, depending on package complexity, this may take a few seconds...",
        logging.INFO,
    )
    pkg_finder = PackageFinder()
    return pkg_finder.run()
